"""
Core
====

This module contains the core classes and functions of the package.

Examples
--------
>>> import clap
>>>
>>>
>>> class ExampleCLI(clap.Parser):
...     \"\"\"Represents a collection of commands for the Example CLI.\"\"\"
...
...     def __init__(self, *args, **kwargs):
...         super().__init__(
...             help="A command-line tool for managing servers.",
...             epilog="Thank you for using Example CLI!",
...             *args,
...             **kwargs,
...         )
...         # do normal object initialization stuff here
...         ...
...
...     @clap.group()
...     def add(self, *args: Any, **kwargs: Any) -> None:
...         \"\"\"A group of commands for adding things.\"\"\"
...         pass
...
...     @add.command()
...     def user(self, id: int, /, *, admin: bool = False) -> None:
...         \"\"\"Adds a user to the system.
...
...         Parameters
...         ----------
...         id : int
...             The ID of the user to add.
...         admin : bool, default=False
...             Whether the user should be an administrator.
...         \"\"\"
...         ...

"""
from __future__ import annotations

import inspect
import re
import sys
from typing import (
    Any,
    Callable,
    Generic,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypeVar,
    get_type_hints,
)

from .arguments import Argument, DefaultHelp, Option
from .help import HelpBuilder, HelpFormatter, HelpItem
from .metadata import Conflicts, Range, Requires, Short
from .utils import MISSING, fold_text

__all__ = (
    "HasCommands",
    "Command",
    "command",
    "Group",
    "group",
)

T = TypeVar("T")
P = ParamSpec("P")

COMMAND_HELP_REGEX = re.compile(r"^(.*?)(?:\n\n|\Z)", re.DOTALL)
COMMAND_LONG_DESCRIPTION_REGEX = re.compile(
    r"^(?:.*?\n\n)(.*?)(?:\n\n|\Z)", re.DOTALL
)
SECTION_REGEX_FMT = r"{section_name}\n-+\n\s*(.*?)(?:\n\n|\Z)"
PARAMETER_SECTION_REGEX = re.compile(
    SECTION_REGEX_FMT.format(section_name="Parameters"), re.DOTALL
)
OTHER_PARAMETER_SECTION_REGEX = re.compile(
    SECTION_REGEX_FMT.format(section_name="Other Parameters"), re.DOTALL
)
PARAMETER_DESCRIPTION_REGEX = re.compile(
    r"(?P<name>\S+)\s*:.*?\n(?P<description>.*?)(?=\S+\s*:|\Z)", re.DOTALL
)


class Command(Generic[T]):
    """Represents an executable command on the command-line.

    Parameters
    ----------
    callback : Callable
        The function to wrap.

    Attributes
    ----------
    callback : Callable
        The function wrapped by the command.
    name : str
        The name of the command.
    help : str
        The help message for the command.
    arguments : list[Argument[Any]]
        The arguments of the command.
    options : dict[str, Option[Any]]
        The options of the command.
    short_option_map : dict[str, str]
        A mapping of short option names to their full names.
    """

    def __init__(self, callback: Callable[P, T], /, **kwargs: Any) -> None:
        self.parent: HasCommands | None = kwargs.get("parent", None)
        if not callable(callback):
            raise TypeError(
                f"Command callback must be callable, not {type(callback)!r}"
            )
        else:
            self.callback = callback

        self.name: str = kwargs.pop("name") or callback.__name__

        parsed_docstring = _parse_docstring(callback)
        self.help = kwargs.pop("help", parsed_docstring["help"])
        self.arguments: list[Argument[Any]] = kwargs.get("arguments", [])
        self.options: dict[str, Option[Any]] = kwargs.get("options", {})
        self.short_option_map: dict[str, str] = kwargs.get(
            "short_option_map", {}
        )
        self.add_option(DefaultHelp)

        parameters = inspect.signature(callback).parameters
        parsed_descriptions: dict[str, str] = parsed_docstring["parameters"]
        parameter_types = get_type_hints(callback, include_extras=True)
        for parameter in parameters.values():
            if parameter.name == "self":
                continue

            argument = self._create_argument(
                parameter, parsed_descriptions, parameter_types
            )

            if isinstance(argument, Option):
                self.options[argument.name] = argument
            else:
                self.arguments.append(argument)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        if self.parent is not None:
            return self.callback(self.parent, *args, **kwargs)  # type: ignore
        else:
            return self.callback(*args, **kwargs)

    def print_help(self, *, help_fmt: HelpFormatter = HelpFormatter()) -> None:
        """Display the help message for the command.

        Parameters
        ----------
        help_fmt : HelpFormatter, optional
            The help formatter to use for formatting the help message.
        """
        builder = HelpBuilder(formatter=help_fmt)
        builder.add_line(self.help)

        usage = self.name
        if self.options:
            usage += " [OPTIONS]"

        optional_remaining = False
        for argument in self.arguments:
            if not optional_remaining and argument.default is not MISSING:
                optional_remaining = True
            fmt = " <%s>" if not optional_remaining else " [%s]"
            usage += fmt % argument.name.upper()

        builder.add_header("USAGE", usage)

        if self.options:
            builder.add_header("OPTIONS")
            for option in self.options.values():
                builder.add_item(option)

        if self.arguments:
            builder.add_header("ARGUMENTS")
            for argument in self.arguments:
                builder.add_item(argument)

        sys.stdout.write(builder.build())

    def help_item_format(self) -> HelpItem:
        """Implement the :protocol:`HasHelpItemFormat` protocol.

        Returns
        -------
        HelpItem
            The name and description of the command to use in the help message.
        """
        return HelpItem(self.name, self.help)

    def add_option(self, option: Option[Any]) -> None:
        """Add an option to the command.

        Parameters
        ----------
        option : Option[Any]
            The option to add.
        """
        if option.name in self.options:
            raise ValueError(f"Option {option.name!r} already exists.")
        self.options[option.name] = option

        if option.short is not MISSING:
            if option.short in self.short_option_map:
                raise ValueError(
                    f"Short option {option.short!r} already exists."
                )
            self.short_option_map[option.short] = option.name

    def remove_option(self, name: str) -> None:
        """Remove an option from the command.

        Parameters
        ----------
        name : str
            The name of the option to remove.
        """
        if name not in self.options:
            raise ValueError(f"Option {name!r} does not exist.")
        del self.options[name]

        for k, v in self.short_option_map.items():
            if v == name:
                del self.short_option_map[k]
                break

    def add_argument(self, argument: Argument[Any]) -> None:
        """Add an argument to the command.

        Parameters
        ----------
        argument : Argument[Any]
            The argument to add.
        """
        self.arguments.append(argument)

    def remove_argument(self, name: str) -> None:
        """Remove an argument from the command.

        Parameters
        ----------
        name : str
            The name of the argument to remove.
        """
        for i, argument in enumerate(self.arguments):
            if argument.name == name:
                del self.arguments[i]
                break
        else:
            raise ValueError(f"Argument {name!r} does not exist.")

    def _create_argument(
        self,
        parameter: inspect.Parameter,
        descriptions: dict[str, str],
        types: dict[str, type[Any]],
        **attrs: Any,
    ) -> Argument[Any] | Option[Any]:
        """Return a initialized command-line argument based on the given
        parameter in the function's signature.

        Parameters
        ----------
        parameter : inspect.Parameter
            The parameter to create the argument from.
        descriptions : dict[str, str]
            A dictionary mapping the names of the parameters to their
            descriptions.
        attrs : Any
            Additional keyword arguments to pass to the constructor.

        Returns
        -------
        Argument[Any]
            The initialized argument.
        """
        Empty = inspect.Parameter.empty
        data = {
            "name": (name := parameter.name),
            "default": (
                parameter.default
                if parameter.default is not Empty
                else MISSING
            ),
        }

        try:
            data["help"] = attrs.get("help", descriptions[name])
        except KeyError:
            raise ValueError(
                f"Missing documentation for parameter {name!r} of "
                f"function {self.callback.__name__!r}"
            )

        try:
            cls = attrs.get("cls", types[name])

            if cls is inspect.Parameter.empty:
                raise TypeError
        except (KeyError, TypeError) as e:
            raise ValueError(
                f"Missing type annotation for parameter {name!r} of "
                f"function {self.callback.__name__!r}"
            ) from e

        if hasattr(cls, "__metadata__"):
            metadata = cls.__metadata__
            for v in metadata:
                if isinstance(v, Short):
                    data["short"] = v
                elif isinstance(v, Requires):
                    data["requires"] = v
                elif isinstance(v, Conflicts):
                    data["conflicts"] = v
                elif isinstance(v, Range):
                    data["range"] = (v.minimum, v.maximum)
                else:
                    continue  # TODO: Log a warning here.

        if hasattr(cls, "__origin__"):
            cls = cls.__origin__

        data["cls"] = cls

        data.update(attrs)
        return self._get_argument_type(parameter)(**data)

    def _get_argument_type(
        self, parameter: inspect.Parameter, /
    ) -> type[Argument[Any]] | type[Option[Any]]:
        """Factory method for getting the type to use when converting a
        parameter to a command-line object.

        Parameters
        ----------
        parameter : inspect.Parameter
            The parameter to get the type for.

        Returns
        -------
        Argument | Option
            The type to use for the parameter.
        """
        k: TypeAlias = inspect.Parameter
        if parameter.kind in {k.POSITIONAL_ONLY, k.VAR_POSITIONAL}:
            return Argument
        elif parameter.kind in {
            k.POSITIONAL_OR_KEYWORD,
            k.KEYWORD_ONLY,
            k.VAR_KEYWORD,
        }:
            return Option
        else:
            raise NotImplementedError


class Group:
    """Represents a group of related commands.

    Parameters
    ----------
    name : str
        The name of the command group.

    Attributes
    ----------
    name : str
        The name of the command group.
    commands : dict[str, Command]
        The commands in the group.
    """

    def __init__(
        self, callback: Callable[P, T], /, name: str, *args: Any, **kwargs: Any
    ) -> None:
        self.parent: HasCommands | None = kwargs.get("parent", None)
        if not callable(callback):
            raise TypeError(
                f"Command callback must be callable, not {type(callback)!r}"
            )
        else:
            self.callback = callback

        self.name = name or callback.__name__
        self.help = kwargs.pop("help", _parse_docstring(callback)["help"])
        self.commands: dict[str, Command[Any] | Group] = {}
        self.options: dict[str, Option[Any]] = kwargs.get("options", {})
        self.short_option_map: dict[str, str] = kwargs.get(
            "short_option_map", {}
        )
        self.add_option(DefaultHelp)

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> None:
        if args or kwargs:
            raise TypeError("Command groups cannot be called with arguments.")
        else:
            self.print_help()

    def print_help(self, *, help_fmt: HelpFormatter = HelpFormatter()) -> None:
        """Display the help message for the group.

        Parameters
        ----------
        help_fmt : HelpFormatter, optional
            The help formatter to use for formatting the help message.
        """
        builder = HelpBuilder(formatter=help_fmt)
        builder.add_line(self.help)

        usage = self.name
        if self.options:
            usage += " [OPTIONS]"
        if self.commands:
            usage += " <COMMAND> [ARGUMENTS]"

        builder.add_header("USAGE", usage)

        if self.options:
            builder.add_header("OPTIONS")
            for option in self.options.values():
                builder.add_item(option)

        if self.commands:
            builder.add_header("COMMANDS")
            for command in self.commands.values():
                builder.add_item(command)

        sys.stdout.write(builder.build())

    def help_item_format(self) -> HelpItem:
        """Implement the :protocol:`HasHelpItemFormat` protocol.

        Returns
        -------
        HelpItem
            The name and description of the command group to use in the help
            message.
        """
        return HelpItem(self.name, self.help)

    def add_command(self, command: Command[Any] | Group) -> None:
        """Adds a command to the group.

        Parameters
        ----------
        command : Command[Any]
            The command to add.
        """
        if command.name in self.commands:
            raise ValueError(f"Command {command.name!r} already exists.")
        self.commands[command.name] = command

    def remove_command(self, name: str) -> None:
        """Removes a command from the group.

        Parameters
        ----------
        name : str
            The name of the command to remove.
        """
        if name not in self.commands:
            raise ValueError(f"Command {name!r} does not exist.")
        del self.commands[name]

    def add_option(self, option: Option[Any]) -> None:
        """Add an option to the command group.

        Parameters
        ----------
        option : Option[Any]
            The option to add.
        """
        if option.name in self.options:
            raise ValueError(f"Option {option.name!r} already exists.")
        self.options[option.name] = option

        if option.short is not MISSING:
            if option.short in self.short_option_map:
                raise ValueError(
                    f"Short option {option.short!r} already exists."
                )
            self.short_option_map[option.short] = option.name

    def remove_option(self, name: str) -> None:
        """Remove an option from the command group.

        Parameters
        ----------
        name : str
            The name of the option to remove.
        """
        if name == "help":
            raise ValueError("Cannot remove the help option.")

        if name not in self.options:
            raise ValueError(f"Option {name!r} does not exist.")
        del self.options[name]

        for k, v in self.short_option_map.items():
            if v == name:
                del self.short_option_map[k]
                break

    def command(
        self, name: str = MISSING, **attrs: Any
    ) -> Callable[[Callable[P, T]], Command[T]]:
        """A convenience wrapper for registering a command with the group.

        Parameters
        ----------
        name : str, optional
            The name of the command. If not provided, the name of the function
            will be used.
        cls : Type[T], optional
            The class to use for constructing the command. If not provided, the
            default `Command` class will be used.
        attrs : Any
            Additional keyword arguments to pass to the constructor.

        Returns
        -------
        Callable[[Callable[P, T]], T]
            The decorator.
        """

        def wrapper(func: Callable[P, T]) -> Command[T]:
            attrs.setdefault("parent", self.parent or self)
            c = command(name=name, **attrs)(func)
            self.add_command(c)
            return c

        return wrapper

    def group(
        self, name: str = MISSING, **attrs: Any
    ) -> Callable[[Callable[P, None]], Group]:
        """A convenience wrapper for registering a command group.

        Parameters
        ----------
        name : str, optional
            The name of the command group. If not provided, the name of the
            function will be used.
        cls : Type[T], optional
            The class to use for constructing the command group. If not
            provided, the default `Group` class will be used.
        attrs : Any
            Additional keyword arguments to pass to the constructor.

        Returns
        -------
        Callable[[Callable[P, T]], T]
            The decorator.
        """

        def wrapper(func: Callable[P, None]) -> Group:
            attrs.setdefault("parent", self)
            g = group(name=name, **attrs)(func)
            self.add_command(g)
            return g

        return wrapper


class HasCommands(Protocol):
    """Protocol for objects that provide commands."""

    commands: dict[str, Command[Any] | Group]

    def add_command(self, command: Command[Any] | Group, /) -> None:
        """Add a command to the object.

        Parameters
        ----------
        command : Command[Any]
            The command to add.
        """
        ...

    def remove_command(self, name: str, /) -> None:
        """Remove a command from the object.

        Parameters
        ----------
        name : str
            The name of the command to remove.
        """
        ...


def command(
    name: str = MISSING, **attrs: Any
) -> Callable[[Callable[P, T]], Command[T]]:
    """Decorator to register a function as a command.

    Parameters
    ----------
    name : str, optional
        The name of the command. If not provided, the name of the function will
        be used.
    attrs : Any
        Additional keyword arguments to pass to the constructor.

    Returns
    -------
    Callable[[Callable[P, T]], Command[T]]
        The decorator.
    """

    def decorator(callback: Callable[P, T], /) -> Command[T]:
        if isinstance(callback, Command):
            raise TypeError(f"Callback {callback!r} is already a command.")
        return Command(callback, name=name, **attrs)

    return decorator


def group(
    name: str = MISSING, **attrs: Any
) -> Callable[[Callable[P, None]], Group]:
    """Decorator to register a function as a command group.

    Parameters
    ----------
    name : str, optional
        The name of the command group. If not provided, the name of the
        function will be used.
    attrs : Any
        Additional keyword arguments to pass to the constructor.

    Returns
    -------
    Callable[[Callable[P, T]], Group]
        The decorator.
    """

    def decorator(callback: Callable[P, None], /) -> Group:
        if isinstance(callback, Group):
            raise TypeError(
                f"Callback {callback!r} is already a command group."
            )
        return Group(callback, name=name, **attrs)

    return decorator


def _parse_docstring(callback: Callable[..., Any], /) -> dict[str, Any]:
    """Extracts information from a function's docstring for auto-constructing
    commands.

    Parameters
    ----------
    callback : Callable
        The function to extract information from.

    Returns
    -------
    dict[str, Any]
        A dictionary containing the extracted information.
    """
    if (docstring := inspect.getdoc(callback)) is None:
        raise ValueError(f"Function {callback.__name__!r} has no docstring")
    return {
        "help": _extract_command_help(docstring),
        "parameters": _extract_parameter_descriptions(docstring),
    }


def _extract_command_help(
    docstring: str,
    /,
    *,
    pattern: re.Pattern[str] = COMMAND_HELP_REGEX,
) -> str | None:
    """Extract the first line of a command's docstring as the help description.

    Parameters
    ----------
    docstring : str
        The docstring to extract the description from.
    pattern : re.Pattern[str], optional
        The pattern to use for extracting the description.

    Returns
    -------
    str
        The first line of the docstring of the function.
    None
        If the function has no docstring.
    """
    if (match := pattern.match(docstring)) is None:
        return None
    return fold_text(match.group(1))


def _extract_parameter_descriptions(
    docstring: str,
    /,
    *,
    section_pattern: re.Pattern[str] = PARAMETER_SECTION_REGEX,
    description_pattern: re.Pattern[str] = PARAMETER_DESCRIPTION_REGEX,
) -> dict[str, str]:
    """Extract the descriptions of the parameters of a function from its
    docstring.

    Parameters
    ----------
    docstring : str
        The docstring to extract the descriptions from.
    pattern : re.Pattern[str], optional
        The pattern to use for extracting the descriptions.

    Returns
    -------
    dict[str, str]
        A dictionary mapping the names of the parameters to their descriptions.
    """
    if (match := section_pattern.search(docstring)) is None:
        return {}
    return {
        name: fold_text(description)
        for name, description in description_pattern.findall(match.group(1))
    }



"""
CLAP
====

Command Line Argument Parser (CLAP) is a library for parsing command-line
arguments.

Examples
--------
>>> from typing import Annotated
>>>
>>> import clap
>>> from clap.metadata import Short, Conflicts
>>>
>>> @clap.command(name="start")
... def start(
...     server: str,
...     /,
...     *,
...     verbose: Annotated[bool, Short('v'), Conflicts("quiet")] = False,
...     quiet: Annotated[bool, Short('q')] = False,
... ) -> None:
...     \"\"\"Starts the specified server.
...
...     Parameters
...     ----------
...     server : str
...         The name of the server to start.
...     verbose : bool, default=False
...         Whether to print verbose output.
...     quiet : bool, default=False
...         Whether to suppress all output.
...     \"\"\"
...     ...

"""

__version__ = "0.1.1"

from .arguments import *
from .commands import *
from .help import *
from .metadata import *
from .parser import *



"""
Parser
======

This module contains the parser for the command-line arguments.

Examples
--------
>>> from typing import Any, Annotated
>>>
>>> import clap
>>> from clap.metadata import Short, Conflicts
>>>
>>>
>>> class ExampleCLI(clap.Parser):
...     \"\"\"Represents a collection of commands for the Example CLI.\"\"\"
...
...     def __init__(self, *args: Any, **kwargs: Any) -> None:
...         super().__init__(
...             help="A command-line tool for managing servers.",
...             epilog="Thank you for using Example CLI!",
...             *args,
...             **kwargs,
...         )
...         # do normal object initialization stuff here:
...         self.foo = "bar"
...         self.baz = 69
...
...     @clap.command()
...     def start(
...         self,
...         ip: str,
...         port: int = 8080,
...         /,
...         *,
...         verbose: Annotated[bool, Short("v"), Conflicts("quiet")] = False,
...         quiet: Annotated[bool, Short("q"), Conflicts("verbose")] = False,
...     ) -> None:
...         \"\"\"Starts the specified server.
...
...         Parameters
...         ----------
...         ip : str
...             The IP address of the server to start.
...         port : int, default=25565
...             The port of the server to start.
...         verbose : bool, default=False
...             Whether to print verbose output.
...         \"\"\"
...         ...

"""
from __future__ import annotations

import os
import sys
from typing import Any, Iterable, ParamSpec, TypeVar

from .arguments import DefaultHelp, Option
from .commands import Command, Group
from .help import HelpBuilder, HelpFormatter
from .lexer import Lexer, Token, TokenType
from .utils import MISSING

__all__ = ("Parser",)

T = TypeVar("T")
P = ParamSpec("P")


class Parser:
    """Represents the parser for command-line arguments.

    Parameters
    ----------
    help : str
        The help message for the parser.
    epilog : str
        The epilog for the parser.

    Attributes
    ----------
    help : str
        The help message for the parser.
    epilog : str
        The epilog for the parser.
    program : str
        The name of the program.
    commands : dict[str, Command[Any]]
        The commands that the user can execute.
    options : dict[str, Option[Any]]
        The options that the user can pass to the parser.
    short_option_map : dict[str, str]
        A mapping of short option names to long option names.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.help = kwargs.pop("help", MISSING)
        self.epilog = kwargs.pop("epilog", MISSING)
        self.program = os.path.basename(sys.argv[0])
        self.options: dict[str, Option[Any]] = {}
        self.short_option_map: dict[str, str] = {}

        self.commands: dict[str, Command[Any] | Group] = {}
        for command in self.__class__.__dict__.values():
            if isinstance(command, (Command, Group)):
                command.parent = self
                self.commands[command.name] = command

        self.add_option(DefaultHelp)

    def add_command(self, command: Command[Any] | Group, /) -> None:
        """Adds a command to the parser.

        Parameters
        ----------
        command : Command[Any] | Group
            The command to add.
        """
        if not isinstance(command, (Command, Group)):
            raise TypeError(
                f"Expected a command or group, got {type(command).__name__}."
            )
        elif command.name in self.commands:
            raise ValueError(f"Command {command.name} already exists.")
        else:
            self.commands[command.name] = command

    def remove_command(self, name: str, /) -> None:
        """Removes a command from the parser.

        Parameters
        ----------
        name : str
            The name of the command to remove.
        """
        if name not in self.commands:
            raise ValueError(f"Command {name} does not exist.")
        del self.commands[name]

    def add_option(self, option: Option[Any], /) -> None:
        """Adds an option to the parser.

        Parameters
        ----------
        option : Option[Any]
            The option to add.
        """
        if option.short is not MISSING:
            if option.short in self.short_option_map:
                raise ValueError(
                    f"Short option {option.short} already exists for option "
                    f"{self.short_option_map[option.short]}. Cannot add "
                    f"option {option.name}."
                )
            self.short_option_map[option.short] = option.name

        if option.name in self.options:
            raise ValueError(f"Option {option.name} already exists.")
        self.options[option.name] = option

    def remove_option(self, name: str, /) -> None:
        """Removes an option from the parser.

        Parameters
        ----------
        name : str
            The name of the option to remove.
        """
        if name == "help":
            raise ValueError("The help option cannot be removed.")

        if name not in self.options:
            raise ValueError(f"Option {name} does not exist.")
        else:
            del self.options[name]

        for short, long in self.short_option_map.items():
            if long == name:
                del self.short_option_map[short]
                break

    def print_help(
        self,
        *,
        help_fmt: HelpFormatter = HelpFormatter(),
    ) -> None:
        """Prints the help message for the parser.

        Parameters
        ----------
        help_fmt : HelpFormatter, optional
            The help formatter to use. If not provided, the default formatter
            will be used.
        """
        builder = HelpBuilder(formatter=help_fmt)
        builder.add_line(self.help)

        usage = self.program
        if self.commands:
            usage += " <COMMAND> [OPTIONS] [ARGUMENTS]"
        else:
            # There should always be at least the "help" option.
            usage += " [OPTIONS]"
        builder.add_header("USAGE", usage)
        builder.add_line(
            "For more information on a specific command, use "
            "'<COMMAND> --help'."
        )

        builder.add_header("OPTIONS")
        for option in self.options.values():
            builder.add_item(option)

        builder.add_header("COMMANDS")
        for command in self.commands.values():
            builder.add_item(command)
        builder.placeholder = "No commands available."
        builder.flush_buffer()
        builder.placeholder = None

        if self.epilog is not MISSING:
            builder.add_newline()
            builder.add_line(self.epilog)

        sys.stdout.write(builder.build())

    def parse(
        self,
        argv: list[str] = sys.argv,
        /,
        *,
        help_fmt: HelpFormatter = HelpFormatter(),
    ) -> None:
        """Parses the command-line arguments.

        Parameters
        ----------
        argv : list[str], optional
            The command-line arguments to parse. If not provided, the arguments
            passed to the program will be used.
        """
        try:
            (args := argv.copy()).pop(0)
        except IndexError as e:
            raise ValueError("Expected at least the program name.") from e

        if len(args) < 1:
            self.print_help(help_fmt=help_fmt)
            return

        lexer = Lexer(args)
        cursor = lexer.cursor()
        deferred: list[Token] = []
        # For passing around context between various parts of the parser.
        ctx: dict[str, Any] = {
            "command": self,
            "arguments": [],
            "options": {},
        }

        for token in lexer.tokens(cursor, peek=False):
            match token.type:
                # All flags are deferred so that both `clap --help command`
                # and `clap command --help` are equivalent.
                case TokenType.LONG_OPTION:
                    deferred.append(token)
                case TokenType.SHORT_OPTION:
                    deferred.append(token)
                    # TODO: Handle options that take arguments.
                    # I think it would be easiest to have the Lexer take
                    # commands, options, and arguments as parameters and
                    # return the appropriate token type. This would require
                    # some changes to the Lexer, but it would make the parser
                    # much simpler.
                case TokenType.ARGUMENT:
                    self._parse_argument(token, ctx)
                case _:
                    raise ValueError(f"Invalid token type: {token.type}.")

        self._process_deferred(deferred, ctx)

        for option in ctx["command"].options.values():
            name = option.name.replace("-", "_")
            if name in ctx["options"] or option.default is MISSING:
                continue

            ctx["options"][name] = option.default

        if ctx["options"].pop("help", False) is True:
            ctx["command"].print_help(help_fmt=help_fmt)
            return

        ctx["command"](*ctx["arguments"], **ctx["options"])

    def _process_deferred(
        self,
        deferred: list[Token],
        ctx: dict[str, Any],
        /,
    ) -> None:
        """Processes deferred tokens.

        Parameters
        ----------
        deferred : list[Token]
            The list of deferred tokens.
        ctx : dict[str, Any]
            The context of the command.
        """
        while deferred:
            token = deferred.pop(0)
            next_token = (
                deferred[0]
                if deferred and token.type is TokenType.LONG_OPTION
                else None
            )

            match token.type:
                case TokenType.LONG_OPTION:
                    self._parse_option(token, ctx, next_token=next_token)
                case TokenType.SHORT_OPTION:
                    self._parse_short_option(token, ctx, next_token=next_token)
                case _:
                    raise ValueError(f"Invalid token type: {token.type}.")

    def _parse_option(
        self,
        token: Token,
        ctx: dict[str, Any],
        /,
        next_token: Token | None = None,
    ) -> None:
        """Processes an option.

        Parameters
        ----------
        token : Token
            The token to process.
        ctx : dict[str, Any]
            The context of the command.
        next_token : Token | None, optional
            The next token. This is used for options that take arguments.
        """
        assert token.type is TokenType.LONG_OPTION
        name = token.value.lstrip("-")
        value: Any = MISSING
        if "=" in name:
            name, value = name.split("=", maxsplit=1)
        command: Parser | Command[Any] | Group = ctx["command"]
        if name not in command.options:
            raise ValueError(f"Unknown option: {name}.")
        option = command.options[name]

        while value is MISSING:
            if option.cls is bool:
                value = True
            elif next_token is not None:
                if next_token.type is TokenType.ARGUMENT:
                    value = next_token.value
                else:
                    next_token = None
                    # value is still MISSING, so the loop will rerun
                    # and continue checking the remaining conditions.
            elif option.default is not MISSING:
                value = option.default
            else:
                raise ValueError(f"Option '{name}' requires an argument.")

        minimum, maximum = option.range
        if minimum > 1:
            value = value.split(",")
            if maximum is not None and len(value) > maximum:
                raise ValueError(
                    f"Option '{name}' takes up to ({maximum}) arguments, "
                    f"got {len(value)}."
                )
            elif len(value) < minimum:
                raise ValueError(
                    f"Option '{name}' takes at least ({minimum}) arguments, "
                    f"got {len(value)}."
                )

        name = name.replace("-", "_")
        ctx["options"][name] = (
            option.cls(value)
            if not isinstance(value, Iterable) or isinstance(value, str)
            else [option.cls(v) for v in value]
        )

    def _parse_short_option(
        self,
        token: Token,
        ctx: dict[str, Any],
        /,
        next_token: Token | None = None,
    ) -> None:
        """Processes a short option.

        Parameters
        ----------
        token : Token
            The token to process.
        ctx : dict[str, Any]
            The context of the command.
        next_token : Token | None, optional
            The next token. This is used for options that take arguments.
        """
        assert token.type is TokenType.SHORT_OPTION
        command: Parser | Command[Any] | Group = ctx["command"]
        name = token.value.lstrip("-")
        if name not in command.short_option_map:
            raise ValueError(f"Unknown short option: {name}.")
        long_name = command.short_option_map[name]
        self._parse_option(
            Token(TokenType.LONG_OPTION, long_name),
            ctx,
            next_token=next_token,
        )

    def _parse_argument(self, token: Token, ctx: dict[str, Any], /) -> None:
        """Parses an argument.

        Parameters
        ----------
        token : Token
            The token to parse.
        ctx : dict[str, Any]
            The context of the command.
        """
        assert token.type is TokenType.ARGUMENT
        command: Parser | Command[Any] | Group = ctx["command"]
        if isinstance(command, (Parser, Group)):
            try:
                command.commands[token.value]
            except KeyError:
                raise ValueError(f"Unknown command: {token.value}.")

            ctx["command"] = self.commands[token.value]
            return

        index = len(ctx["arguments"])
        try:
            argument = ctx["command"].arguments[index]
        except IndexError:
            cls = str
        else:
            cls = argument.cls
        finally:
            ctx["arguments"].append(cls(token.value))



"""
Help
====

A collection of functions and classes for generating the help message.

"""
from __future__ import annotations

import textwrap
from typing import NamedTuple, Protocol

__all__ = [
    "HelpFormatter",
]


class HelpItem(NamedTuple):
    """Represents an item in the help message."""

    name: str
    help: str


class HasHelpItemFormat(Protocol):
    """A protocol for objects that can be formatted as a `HelpItem`."""

    def help_item_format(self) -> HelpItem:
        ...


class HelpFormatter:
    """A class for formatting help messages.

    Parameters
    ----------
    width : int, optional
        The maximum width of the help message.
    name_width : int, optional
        The width of the name column.
    indent : int, optional
        The number of spaces to indent each item under a header.
    compact : bool, optional
        Whether to omit newlines between sections.
    """

    def __init__(
        self,
        width: int = 80,
        name_width: int = 20,
        indent: int = 2,
        compact: bool = False,
    ) -> None:
        if width < 50:
            raise ValueError(f"width must be at least 50, got {width}")
        self.width = width
        self.name_width = name_width
        self.indent = indent
        self.compact = compact


class HelpBuilder:
    """Represents the help message for a command-line interface."""

    def __init__(
        self,
        formatter: HelpFormatter = HelpFormatter(),
    ) -> None:
        """A builder class for generating the help message text.

        Parameters
        ----------
        width : int, optional
            The maximum width of the help message.
        name_width : int, optional
            The width of the name column.
        indent : int, optional
            The number of spaces to indent each item under a header.
        compact : bool, optional
            Whether to omit newlines between sections.
        """
        self.fmt = formatter
        self._text = ""
        self._buffer: list[HelpItem] = []
        self._placeholder: str | None = None  # see `placeholder` property

    @property
    def placeholder(self) -> str | None:
        """Text to display when there are no items in a section
        (e.g. "No commands available." or "No options available.").
        """
        return self._placeholder

    @placeholder.setter
    def placeholder(self, value: str | None) -> None:
        self._placeholder = value

    def build(self) -> str:
        """Builds the help message."""
        self.flush_buffer()
        return self._text

    def add_header(self, header: str, description: str = "") -> None:
        """Add a new section to the help message.

        Parameters
        ----------
        header : str
            The name of the section.
        description : str, optional
            A brief description of the section.
        """
        self.flush_buffer()
        if not self.fmt.compact:
            self.add_newline()
        _description = textwrap.wrap(
            description,
            # -2 for the colon and space
            width=self.fmt.width - len(header) - 2,
            max_lines=1,
        )
        description = _description.pop(0) if _description else description
        self.add_line(f"{header}: {description}")

    def add_item(self, item: HasHelpItemFormat) -> None:
        """Add an item to the help message.

        Parameters
        ----------
        item : HasHelpItemFormat
            The item to add.
        """
        self._buffer.append(item.help_item_format())

    def add_line(self, line: str) -> None:
        """Add a line to the help message.

        Parameters
        ----------
        line : str
            The line to add.
        """
        self._text += line + "\n"

    def add_newline(self, force: bool = False) -> None:
        """Add a newline to the help message."""
        self.add_line("") if force or not self.fmt.compact else None

    def flush_buffer(self) -> None:
        """Writes all of the items in the buffer to the help message. This is
        typically called automatically when a new section is being added or
        the message is being built.
        """
        indent = " " * self.fmt.indent
        if len(self._buffer) < 1:
            if self.placeholder is not None:
                self.add_line(f"{indent}{self.placeholder}")
            return
        # Perform calculations with static values outside of the while loop
        # to avoid needlessly repeating operations on each iteration.
        longest_name = max(len(name) for name, _ in self._buffer)
        # +2 for the two spaces between the name and description
        subsequent_indent = " " * (self.fmt.indent + longest_name + 2)
        width = self.fmt.width - self.fmt.indent - longest_name - 2
        while len(self._buffer) > 0:
            item = self._buffer.pop(0)
            description = "\n".join(
                textwrap.wrap(
                    item.help,
                    width=width,
                    subsequent_indent=subsequent_indent,
                    max_lines=2,
                )
            )
            self.add_line(
                f"{indent}{item.name.ljust(longest_name)}  {description}"
            )



"""
Basic
=====

This example demonstrates the basic usage of CLAP.

"""
import clap


class MyCommands(clap.Parser):
    """Represents the commands that are available to the user."""

    def __init__(self) -> None:
        super().__init__(
            help="This is an example CLI tool created using CLAP.",
            epilog="Thank you for using CLAP!",
        )

    @clap.command()
    def greet(self, name: str, /, nervous: bool = False) -> None:
        """Prints a greeting to the specified name.

        Parameters
        ----------
        name : str
            The name of the person to greet.
        nervous : bool, default=False
            Whether to greet the person nervously.
        """
        if nervous is True:
            print(f"Um... hello, {name}...")
        else:
            print(f"Hello, {name}!")


if __name__ == "__main__":
    parser = MyCommands()
    parser.parse()


"""
Output using [ndg.clap 0.1.0]:
-----------------------------------------------------------------------------
$ python examples/basic.py --help
This is an example CLI tool created using CLAP.

USAGE: basic.py <COMMAND> [OPTIONS] [ARGUMENTS]
For more information on a specific command, use '<COMMAND> --help'.

OPTIONS:
  -h, --help  Display this help message and exit. [default: False]

COMMANDS:
  greet  Prints a greeting to the specified name.

Thank you for using CLAP!

$ python examples/basic.py greet "Gojo Satoru"
$ python examples/basic.py greet --help
Prints a greeting to the specified name.

USAGE: greet [OPTIONS] <NAME>

OPTIONS:
  -h, --help   Display this help message and exit. [default: False]
  --nervous  Whether to greet the person nervously. [default: False]

ARGUMENTS:
  name  The name of the person to greet. (required)

Hello, Gojo Satoru!

$ python examples/basic.py greet --nervous "Gojo Satoru"
Um... hello, Gojo Satoru...
"""



import unittest
from typing import Annotated, Any

import clap
from clap.metadata import Conflicts, Short


def test_argument(argument: clap.Argument, expected: dict[str, Any]) -> None:
    """Tests an argument."""
    assert argument.name == expected["name"]
    assert argument.help == expected["help"]
    assert argument.cls == expected["cls"]
    assert argument.default == expected["default"]
    assert argument.range == expected["range"]


def test_option(option: clap.Option, expected: dict[str, Any]) -> None:
    """Tests an option."""
    assert option.name == expected["name"]
    assert option.help == expected["help"]
    assert option.cls == expected["cls"]
    assert option.default == expected["default"]
    assert option.range == expected["range"]
    assert option.short == expected["short"]
    assert option.conflicts == expected["conflicts"]
    assert option.requires == expected["requires"]


def test_short_option_map(
    short_option_map: dict[str, str], expected: dict[str, Any]
) -> None:
    """Tests a short option map."""
    for short, long in short_option_map.items():
        assert short in expected["short_option_map"].keys()
        assert long in expected["short_option_map"].values()


def test_command(command: clap.Command, test: dict[str, Any]) -> None:
    """Tests a command."""
    assert command.name == test["name"]
    assert command.help == test["help"]
    for argument, expected_a in zip(command.arguments, test["arguments"]):
        test_argument(argument, expected_a)
    for option, expected_o in zip(command.options, test["options"]):
        test_option(option, expected_o)
    assert command.short_option_map == test["short_option_map"]


class Test(clap.Parser):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(
            help="A command-line tool for managing servers.",
            epilog="Thank you for using Test!",
            *args,
            **kwargs,
        )

    @clap.command()
    def start(self, server: str, /) -> None:
        """Starts the specified server.

        Parameters
        ----------
        server : str
            The name of the server to start.
        """
        pass

    @clap.command()
    def start_default(self, /, auto_reconnect: bool = False) -> None:
        """Starts the default server.

        Parameters
        ----------
        auto_reconnect : bool, default=False
            Whether to automatically reconnect to the server if it stops.
        """
        pass

    @clap.command(name="stop")
    def not_go(self, server: str, /) -> None:
        """Stops the specified server.

        Parameters
        ----------
        server : str
            The name of the server to stop.
        """
        pass

    @clap.command()
    def setup(
        self,
        /,
        quiet: Annotated[bool, Short("q"), Conflicts("verbose")] = False,
        verbose: Annotated[bool, Short("v"), Conflicts("quiet")] = False,
    ) -> None:
        """Sets up the specified server.

        Parameters
        ----------
        quiet : bool, default=False
            Whether to suppress all output.
        verbose : bool, default=False
            Whether to print verbose output.
        """
        pass


class TestCommandDecorator(unittest.TestCase):
    def test_arguments(self):
        expected = {
            "start": {
                "name": "start",
                "help": "Starts the specified server.",
                "arguments": [
                    {
                        "name": "server",
                        "help": "The name of the server to start.",
                        "cls": str,
                        "default": clap.utils.MISSING,
                        "range": clap.metadata.Range(1, 1),
                    },
                ],
                "options": {},
                "short_option_map": {},
            },
        }

        t = Test()
        for command in t.commands.values():
            test_command(command, expected[command.name])



"""
Utils
=====

This module contains utility functions and classes used by the other modules.

"""
from __future__ import annotations

import re
from typing import Any


class _Missing:
    __slots__ = ()

    def __eq__(self, other: Any) -> bool:
        return False

    def __bool__(self) -> bool:
        return False

    def __hash__(self) -> int:
        return 0

    def __repr__(self) -> str:
        return "..."


MISSING: Any = _Missing()


def fold_text(text: str, /) -> str:
    """Remove unnecessary whitespaces and replace tabs and newlines with
    a single space.

    Parameters
    ----------
    text : str
        The text to clean.

    Returns
    -------
    str
        The cleaned text.
    """
    return re.sub(r"\s+", " ", text).strip()



"""
Metadata
========

This module contains helper types for use with `typing.Annotated` to provide
additional information about a command's arguments and options.

Examples
--------
>>> from typing import Annotated
>>>
>>> import clap
>>> from clap.metadata import Short, Conflicts
>>>
>>>
>>> @clap.command(name="start")
... def start(
...     server: str,
...     /,
...     *,
...     verbose: Annotated[bool, Short('v'), Conflicts("quiet")] = False,
...     quiet: Annotated[bool, Short('q'), Conflicts("verbose")] = False,
... ) -> None:
...     \"\"\"Starts the specified server.
...
...     Parameters
...     ----------
...     server : str
...         The name of the server to start.
...     verbose : bool, default=False
...         Whether to print verbose output.
...     quiet : bool, default=False
...         Whether to suppress all output.
...
...     \"\"\"
...     ...

"""
from __future__ import annotations

from typing import NamedTuple

__all__ = (
    "Range",
    "Short",
    "Requires",
    "Conflicts",
)


class Short(str):
    """Represents a short name for an option."""

    def __new__(cls, value: str) -> Short:
        if len(value) != 1:
            raise ValueError("Short name must be a single character.")
        return super().__new__(cls, value)


class Requires(set[str]):
    """Represents the names of options that are required by another option."""

    def __new__(cls, *args: str) -> Requires:
        return super().__new__(cls)

    def __init__(self, *args: str) -> None:
        super().__init__(args)


class Conflicts(set[str]):
    """Represents the names of options that conflict with another option."""

    def __new__(cls, *args: str) -> Conflicts:
        return super().__new__(cls)

    def __init__(self, *args: str) -> None:
        super().__init__(args)


class Range(NamedTuple):
    """Represents a range of values."""

    minimum: int
    maximum: int



"""
Lexer
=====

The lexer is responsible for converting the raw command-line arguments into a
list of tokens that can be parsed by the parser.

"""
from __future__ import annotations

import enum
from typing import Iterable, NamedTuple


class TokenType(enum.IntEnum):
    """Represents the type of token."""

    # A long option (e.g. `--help`).
    LONG_OPTION = enum.auto()
    # A short option (e.g. `-h`).
    SHORT_OPTION = enum.auto()
    # An argument (e.g. `foo`).
    ARGUMENT = enum.auto()


class Token(NamedTuple):
    """Represents a token."""

    type: TokenType
    value: str


class Cursor:
    def __init__(self, begin: int = 0, *, end: int) -> None:
        self._begin = begin
        self._end = end
        self._position = begin

    @property
    def begin(self) -> int:
        return self._begin

    @property
    def end(self) -> int:
        return self._end

    @property
    def position(self) -> int:
        return self._position

    @position.setter
    def position(self, value: int) -> None:
        if not (self.begin <= value <= self.end):
            raise IndexError(
                f"Position not in range {self.begin}..{self.end}: {value}."
            )
        self._position = value


class Lexer:
    def __init__(self, __args: list[str], /) -> None:
        self._args = __args.copy()

    @property
    def escape(self) -> int:
        return self._args.index("--") if "--" in self._args else self.end

    @property
    def begin(self) -> int:
        return 0

    @property
    def end(self) -> int:
        return len(self._args)

    def cursor(self) -> Cursor:
        return Cursor(begin=self.begin, end=self.end)

    def tokens(self, cursor: Cursor, /, peek: bool = False) -> Iterable[Token]:
        while cursor.position < self.end:
            argument = self.get_raw_argument(cursor)

            if cursor.position > self.escape:
                yield Token(TokenType.ARGUMENT, argument)
            elif argument.startswith("--"):
                yield Token(TokenType.LONG_OPTION, argument)
            elif argument.startswith("-"):
                for option in argument[1:]:
                    yield Token(TokenType.SHORT_OPTION, option)
            else:
                yield Token(TokenType.ARGUMENT, argument)

            self.advance(cursor) if not peek else None

    def get_raw_argument(self, cursor: Cursor) -> str:
        return self._args[cursor.position]

    def advance(self, cursor: Cursor, n: int = 1) -> None:
        cursor.position += n

    def retreat(self, cursor: Cursor, n: int = 1) -> None:
        self.advance(cursor, -n)

    def seek(self, cursor: Cursor, position: int) -> None:
        self.advance(cursor, position - cursor.position)

    def valid_cursor(self, cursor: Cursor) -> bool:
        return (
            cursor.begin >= self.begin
            and cursor.end <= self.end
            and cursor.begin <= cursor.position <= cursor.end
        )



"""
Argument
========

Represents an argument to a command.

"""
from __future__ import annotations

import functools
from typing import Any, Generic, TypeVar

from .help import HelpItem
from .metadata import Range
from .utils import MISSING

__all__ = (
    "Argument",
    "Option",
    "Switch",
)

T = TypeVar("T")


class Argument(Generic[T]):
    """Represents an argument to a command.

    Parameters
    ----------
    name : str
        The name of the argument.
    help : str
        The help message for the argument.
    cls : type[T]
        The type of the argument.
    default : T | None
        The default value of the argument.
    range : tuple[int | None, int | None]
        The range of values that can be passed to the argument. If the first
        or second element is None, then there is no lower or upper bound,
        respectively.
    """

    def __init__(self, name: str, **kwargs: Any) -> None:
        self.name = name
        self.help: str = kwargs.pop("help", MISSING)
        if not self.help:
            raise ValueError(f"Missing help for argument {name!r}.")
        self.cls: type[T] = kwargs.pop("cls", str)
        self.default: T = kwargs.pop("default", MISSING)
        self.range: Range = kwargs.pop("range", Range(0, 1))

    def __call__(self, value: Any, *args: Any, **kwargs: Any) -> T:
        if value is MISSING:
            if self.default is MISSING:
                raise ValueError(f"Missing value for argument {self.name!r}.")
            else:
                return self.default
        else:
            return self.cls(value)  # type: ignore

    def help_item_format(self) -> HelpItem:
        description = self.help
        if self.default is not MISSING:
            description += f" [default: {self.default!r}]"
        else:
            description += " (required)"
        return HelpItem(self.name, description)


class Option(Argument[T]):
    """Represents an option to a command.

    Parameters
    ----------
    name : str
        The name of the option.
    help : str
        The help message for the option.
    cls : type[T]
        The type of the option.
    default : Optional[T]
        The default value of the option.
    short : str, optional
        A single character alias for the option.
    requires : set[str]
        The names of the options that this option requires.
    conflicts : set[str]
        The names of the options that this option conflicts with.
    """

    def __init__(self, name: str, **kwargs: Any) -> None:
        name = name.replace("_", "-")
        super().__init__(name, **kwargs)
        self.short: str = kwargs.pop("short", MISSING)
        self.requires: set[str] = kwargs.pop("requires", set())
        self.conflicts: set[str] = kwargs.pop("conflicts", set())

    def help_item_format(self) -> HelpItem:
        name = ""
        if self.short is not MISSING:
            name += f"-{self.short}, "
        name += f"--{self.name}"

        description = self.help
        if self.default is not MISSING:
            description += f" [default: {self.default!r}]"
        elif self.cls is bool:
            description += " [default: False]"
        else:
            description += " (required)"

        return HelpItem(name, description)


Switch = functools.partial(Option, cls=bool, default=False, range=(0, 1))
DefaultHelp = Switch(
    name="help",
    help="Display this help message and exit.",
    short="h",
)
